# http_handler.py

import os
import time
import json
import requests
import urllib.parse
import http.server

from config import AUTH_PASSWORD
from device_manager import (
    register_device,
    update_ping,
    device_exists,
    get_device_ip,
    list_device_names
)
from ws_server import broadcast_device_list

class RequestHandler(http.server.SimpleHTTPRequestHandler):
    """
    Threaded HTTP handler for all routes:
    - GET /
    - POST /register
    - POST /ping
    - GET /deviceinfo?device=...
    - GET /readfile?device=X&filename=Y
    - POST /savefile
    - POST /deletefile
    - POST /rebootdevice
    """
    def do_GET(self):
        if self.path == "/":
            self.serve_index()
            return

        if self.path.startswith("/deviceinfo"):
            self.handle_device_info()
            return

        if self.path.startswith("/readfile"):
            self.handle_read_file()
            return

        # If not recognized, fall back to super() which might 404 or attempt to serve a file
        super().do_GET()

    def do_POST(self):
        if self.path == "/register":
            self.handle_register()
            return
        if self.path == "/ping":
            self.handle_ping()
            return
        if self.path == "/savefile":
            self.handle_save_file()
            return
        if self.path == "/deletefile":
            self.handle_delete_file()
            return
        if self.path == "/rebootdevice":
            self.handle_reboot_device()
            return

        self.send_error(404, "Unknown POST endpoint")

    ###########################################################################
    #                              GET /                                      #
    ###########################################################################
    def serve_index(self):
        """
        Serves templates/index.html if present.
        """
        try:
            base_path = os.path.dirname(__file__)
            # You'll need to modify this to point to where you have index.html on your backend server
            template_path = os.path.join(os.path.dirname(base_path), "/home/zimablade/DeviceWrangler/templates", "index.html")
            with open(template_path, "rb") as f:
                content = f.read()

            self.send_response(200)
            self.send_header("Content-Type", "text/html; charset=utf-8")
            self.send_header("Content-Length", str(len(content)))
            self.end_headers()
            self.wfile.write(content)

        except Exception as e:
            self.send_error(500, f"Error loading index.html: {e}")

    ###########################################################################
    #                          POST /register                                  #
    ###########################################################################
    def handle_register(self):
        length = self._content_length()
        if length is None:
            return

        body = self.rfile.read(length)
        parsed = urllib.parse.parse_qs(body.decode("utf-8"))
        dev_name = parsed.get("Device_Name", [""])[0]
        auth_val = parsed.get("Auth", [""])[0]
        ip_val = parsed.get("IP", [""])[0] if "IP" in parsed else self.client_address[0]

        if not dev_name:
            self.send_error(400, "No Device_Name")
            return
        if auth_val != AUTH_PASSWORD:
            self.send_error(403, "Invalid auth")
            return

        register_device(dev_name, ip_val)
        self.send_response(200)
        self.end_headers()
        self.wfile.write(b"Device registered")

        broadcast_device_list()

    ###########################################################################
    #                          POST /ping                                      #
    ###########################################################################
    def handle_ping(self):
        length = self._content_length()
        if length is None:
            return

        body = self.rfile.read(length)
        parsed = urllib.parse.parse_qs(body.decode("utf-8"))
        dev_name = parsed.get("Device_Name", [""])[0]
        auth_val = parsed.get("Auth", [""])[0]
        ip_val = parsed.get("IP", [""])[0] if "IP" in parsed else self.client_address[0]

        if not dev_name or auth_val != AUTH_PASSWORD:
            self.send_error(403)
            return

        update_ping(dev_name, ip_val)
        self.send_response(200)
        self.end_headers()
        self.wfile.write(b"Ping OK")

    ###########################################################################
    #                       GET /deviceinfo?device=...                         #
    ###########################################################################
    def handle_device_info(self):
        parsed = urllib.parse.urlparse(self.path)
        qs = urllib.parse.parse_qs(parsed.query)
        dev_name = qs.get("device", [""])[0]

        if not dev_name or not device_exists(dev_name):
            self.send_error(404, "Unknown device")
            return

        ip = get_device_ip(dev_name)
        url = f"http://{ip}:8080/list"

        try:
            print(f"[DeviceInfo] Fetching file list from {url}")
            r = requests.get(url, timeout=15)
            if r.status_code == 200:
                files = r.json().get("files", [])
                data = json.dumps({"files": files}).encode("utf-8")
                self.send_response(200)
                self.send_header("Content-Type", "application/json")
                self.send_header("Content-Length", str(len(data)))
                self.end_headers()
                self.wfile.write(data)
            else:
                self.send_error(502, f"Device responded {r.status_code}")
        except Exception as e:
            self.send_error(502, f"Error contacting device: {e}")

    ###########################################################################
    #                    GET /readfile?device=...&filename=...                #
    ###########################################################################
    def handle_read_file(self):
        parsed = urllib.parse.urlparse(self.path)
        qs = urllib.parse.parse_qs(parsed.query)
        dev_name = qs.get("device", [""])[0]
        filename = qs.get("filename", [""])[0]

        if not dev_name or not filename:
            self.send_error(400, "Missing device or filename")
            return
        if not device_exists(dev_name):
            self.send_error(404, "Unknown device")
            return

        ip = get_device_ip(dev_name)
        url = f"http://{ip}:8080/file?name={urllib.parse.quote(filename)}"
        print(f"[ReadFile] Device: {dev_name}, file: {filename}, URL: {url}")

        try:
            with requests.get(url, timeout=30, stream=True) as resp:
                if resp.status_code != 200:
                    self.send_error(502, f"Device responded {resp.status_code}")
                    return

                # Start response with unknown length
                self.send_response(200)
                self.send_header("Content-Type", "text/plain; charset=utf-8")
                self.end_headers()

                total_bytes = 0
                for chunk in resp.iter_content(chunk_size=512):
                    if not chunk:
                        break
                    total_bytes += len(chunk)
                    print(f"[ReadFile] Forwarded {len(chunk)} bytes (total {total_bytes})")
                    self.wfile.write(chunk)

                print(f"[ReadFile] Completed {total_bytes} bytes transfer.")
        except Exception as e:
            self.send_error(502, f"Error contacting device: {e}")

    ###########################################################################
    #                        POST /savefile                                    #
    ###########################################################################
    def handle_save_file(self):
        length = self._content_length()
        if length is None:
            return

        body = self.rfile.read(length)
        try:
            payload = json.loads(body.decode("utf-8"))
        except:
            self.send_error(400, "Invalid JSON")
            return

        dev_name = payload.get("device")
        fname = payload.get("filename")
        content = payload.get("content", "")

        if not dev_name or not fname:
            self.send_error(400, "Missing device or filename")
            return
        if not device_exists(dev_name):
            self.send_error(404, "Unknown device")
            return

        ip = get_device_ip(dev_name)
        url = f"http://{ip}:8080/file"
        print(f"[SaveFile] POST {url} => {fname}")

        try:
            r = requests.post(url, json={"name": fname, "content": content}, timeout=15)
            if r.status_code == 200:
                self.send_response(200)
                self.end_headers()
                self.wfile.write(b"File saved successfully.")
            else:
                self.send_error(502, f"Device responded {r.status_code}")
        except Exception as e:
            self.send_error(502, f"Error contacting device: {e}")

    ###########################################################################
    #                       POST /deletefile                                   #
    ###########################################################################
    def handle_delete_file(self):
        length = self._content_length()
        if length is None:
            return

        body = self.rfile.read(length)
        try:
            payload = json.loads(body.decode("utf-8"))
        except:
            self.send_error(400, "Invalid JSON")
            return

        dev_name = payload.get("device")
        fname = payload.get("filename")
        if not dev_name or not fname:
            self.send_error(400, "Missing device or filename")
            return
        if not device_exists(dev_name):
            self.send_error(404, "Unknown device")
            return

        ip = get_device_ip(dev_name)
        url = f"http://{ip}:8080/file?name={urllib.parse.quote(fname)}"
        print(f"[DeleteFile] DELETE {url}")

        try:
            r = requests.delete(url, timeout=15)
            if r.status_code == 200:
                self.send_response(200)
                self.end_headers()
                self.wfile.write(b"File deleted.")
            else:
                self.send_error(502, f"Device responded {r.status_code}")
        except Exception as e:
            self.send_error(502, f"Error contacting device: {e}")

    ###########################################################################
    #                       POST /rebootdevice                                #
    ###########################################################################
    def handle_reboot_device(self):
        length = self._content_length()
        if length is None:
            return

        body = self.rfile.read(length)
        try:
            payload = json.loads(body.decode("utf-8"))
        except:
            self.send_error(400, "Invalid JSON")
            return

        dev_name = payload.get("device")
        if not dev_name or not device_exists(dev_name):
            self.send_error(404, "Unknown device")
            return

        ip = get_device_ip(dev_name)
        url = f"http://{ip}:8080/reboot"
        print(f"[Reboot] POST {url}")

        try:
            r = requests.post(url, timeout=15)
            if r.status_code == 200:
                self.send_response(200)
                self.end_headers()
                self.wfile.write(b"Reboot command sent successfully.")
            else:
                self.send_error(502, f"Device responded {r.status_code}")
        except Exception as e:
            self.send_error(502, f"Error contacting device: {e}")

    ###########################################################################
    #                          Helper for length                               #
    ###########################################################################
    def _content_length(self):
        length_str = self.headers.get("Content-Length")
        if not length_str:
            self.send_error(400, "Missing Content-Length")
            return None
        try:
            return int(length_str)
        except:
            self.send_error(400, "Invalid Content-Length")
            return None
