import usocket as socket
import time
import machine
import os
import ujson
import ure

def StartHttpServer():
    """
    HTTP server on port 8080:
      - GET /list                    => JSON list of files
      - GET /file?name=<fname>       => Return file content (chunked)
      - POST /file                   => JSON: { name, content }
      - DELETE /file?name=<fname>    => Remove file
      - POST /reboot                 => Reboot (with ACK first)
    """
    addr_info = socket.getaddrinfo("0.0.0.0", 8080)[0][-1]
    s = socket.socket()
    s.bind(addr_info)
    s.listen(5)
    print("[HTTP] Listening on port 8080")

    while True:
        try:
            cl, remote_addr = s.accept()
        except Exception as e:
            print("[HTTP] Socket accept error:", e)
            continue

        try:
            HandleClientRequest(cl, remote_addr)
        except Exception as e:
            print("[HTTP] Error while handling request:", e)
        finally:
            cl.close()

def HandleClientRequest(cl, remote_addr):
    cl_file = cl.makefile("rwb", 0)
    request_line = cl_file.readline()
    if not request_line:
        return

    # Parse method & path
    try:
        req_str = request_line.decode("utf-8").strip()
        method, path, _ = req_str.split()
    except:
        SendResponse(cl, 400, b"Bad Request")
        return

    # Read headers (until blank line)
    headers = {}
    while True:
        line = cl_file.readline()
        if not line or line == b"\r\n":
            break
        line_str = line.decode("utf-8").strip()
        key, val = line_str.split(":", 1)
        headers[key.strip()] = val.strip()

    # Possibly read body
    length = int(headers.get("Content-Length", 0))
    body = b""
    if length > 0:
        body = cl_file.read(length)

    print("[HTTP]", method, path, "from", remote_addr)

    # 1) GET /list => Return JSON list of files
    if method == "GET" and path == "/list":
        resp = HandleListFiles()
        SendResponse(cl, 200, resp, "application/json")
        return

    # 2) GET /file?name=<fname> => chunk-based read
    if method == "GET" and path.startswith("/file?name="):
        fname = ExtractQueryParam(path, "name")
        if not fname:
            SendResponse(cl, 400, b"Missing filename")
            return

        if not file_exists(fname):
            SendResponse(cl, 404, b"File not found")
            return

        # Chunk-based approach
        try:
            fsize = os.stat(fname)[6]  # file size
            SendHeaders(cl, 200, "text/plain", fsize)
            with open(fname, "rb") as f:
                while True:
                    chunk = f.read(512)
                    if not chunk:
                        break
                    cl.send(chunk)
        except MemoryError:
            SendResponse(cl, 500, b"MemoryError during file read")
        except Exception as e:
            print("[HTTP] File read error:", e)
            # Can't fully recover mid-stream if we've already sent headers
        return

    # 3) POST /file => JSON { name, content } => write file
    if method == "POST" and path == "/file":
        try:
            data = ujson.loads(body)
            filename = data.get("name", "")
            fcontent = data.get("content", "")
            if not filename:
                SendResponse(cl, 400, "No filename provided")
                return
            ok = HandleWriteFile(filename, fcontent)
            if ok:
                SendResponse(cl, 200, "OK")
            else:
                SendResponse(cl, 500, "File write error")
        except Exception as ex:
            print("[HTTP] JSON parse error:", ex)
            SendResponse(cl, 400, "Invalid JSON")
        return

    # 4) DELETE /file?name=<fname>
    if method == "DELETE" and path.startswith("/file?name="):
        fname = ExtractQueryParam(path, "name")
        if not fname:
            SendResponse(cl, 400, "Missing filename")
            return
        ok = HandleDeleteFile(fname)
        if ok:
            SendResponse(cl, 200, "Deleted")
        else:
            SendResponse(cl, 404, "File not found or delete error")
        return

    # 5) POST /reboot => Reboot the device
    if method == "POST" and path == "/reboot":
        # Acknowledge first
        SendResponse(cl, 200, "ACK: Rebooting\n")
        cl.close()
        time.sleep(1)
        machine.reset()
        return

    # Otherwise: Not found
    SendResponse(cl, 404, b"Not Found")

########################################################################
#                  Helper: Send only HTTP headers first               #
########################################################################

def SendHeaders(cl, status_code, content_type="text/plain", content_length=0):
    status_text = {200: "OK", 400: "Bad Request", 404: "Not Found", 500: "Internal Server Error"}\
        .get(status_code, "OK")
    header = "HTTP/1.0 {} {}\r\n".format(status_code, status_text)
    header += "Content-Type: {}\r\n".format(content_type)
    header += "Content-Length: {}\r\n\r\n".format(content_length)
    cl.send(header.encode("utf-8"))

########################################################################
#                   Helper: Send small response at once               #
########################################################################

def SendResponse(cl, status_code, content, content_type="text/plain"):
    if isinstance(content, str):
        content = content.encode("utf-8")
    c_len = len(content)
    status_text = {200: "OK", 400: "Bad Request", 404: "Not Found", 500: "Internal Server Error"}\
        .get(status_code, "OK")
    header = "HTTP/1.0 {} {}\r\n".format(status_code, status_text)
    header += "Content-Type: {}\r\n".format(content_type)
    header += "Content-Length: {}\r\n\r\n".format(c_len)
    cl.send(header.encode("utf-8"))
    cl.send(content)

########################################################################
#                           FILE OPERATIONS                            #
########################################################################

def HandleListFiles():
    try:
        listing = os.listdir()
        return ujson.dumps({"files": listing})
    except Exception as e:
        print("[ListFiles] Error:", e)
        return ujson.dumps({"files": []})

def HandleWriteFile(fname, content):
    try:
        with open(fname, "w") as f:
            f.write(content)
        return True
    except Exception as ex:
        print("[WriteFile] Error writing:", fname, ex)
        return False

def HandleDeleteFile(fname):
    try:
        os.remove(fname)
        return True
    except Exception as ex:
        print("[DeleteFile] Error removing:", fname, ex)
        return False

def file_exists(fname):
    try:
        return fname in os.listdir()
    except:
        return False

########################################################################
#                         UTILITY FOR PARSING                          #
########################################################################

def ExtractQueryParam(path, key):
    # E.g.: /file?name=main.py
    parts = path.split('?', 1)
    if len(parts) < 2:
        return None
    query = parts[1]
    pattern = key + "="
    idx = query.find(pattern)
    if idx < 0:
        return None
    start = idx + len(pattern)
    end = query.find("&", start)
    if end < 0:
        end = len(query)
    raw_value = query[start:end]
    return urllib_unquote(raw_value)

def urllib_unquote(s):
    s = s.replace('+', ' ')
    i = 0
    out = ""
    while i < len(s):
        c = s[i]
        if c == '%' and i+2 < len(s):
            try:
                val = int(s[i+1:i+3], 16)
                out += chr(val)
                i += 3
                continue
            except:
                pass
        out += c
        i += 1
    return out